@@ -25,7 +25,11 @@
 #ifndef TESTING_GEBSR2GEBSC_HPP
 #define TESTING_GEBSR2GEBSC_HPP
 
+#include "display.hpp"
+#include "flops.hpp"
+#include "gbyte.hpp"
 #include "hipsparse.hpp"
+#include "hipsparse_arguments.hpp"
 #include "hipsparse_test_unique_ptr.hpp"
 #include "unit.hpp"
 #include "utility.hpp"
@@ -71,18 +75,11 @@ void testing_gebsr2gebsc_bad_arg(void)
     int* bsc_col_ptr = (int*)bsc_col_ptr_managed.get();
     T*   bsc_val     = (T*)bsc_val_managed.get();
 
-    if(!bsr_row_ptr || !bsr_col_ind || !bsr_val || !bsc_row_ind || !bsc_col_ptr || !bsc_val)
-    {
-        PRINT_IF_HIP_ERROR(hipErrorOutOfMemory);
-        return;
-    }
-    { //
-        int local_ptr[2] = {0, 1};
-        CHECK_HIP_ERROR(hipMemcpy(
-            bsr_row_ptr, local_ptr, sizeof(int) * (safe_size + 1), hipMemcpyHostToDevice));
-        CHECK_HIP_ERROR(hipMemcpy(
-            bsc_col_ptr, local_ptr, sizeof(int) * (safe_size + 1), hipMemcpyHostToDevice));
-    } //
+    int local_ptr[2] = {0, 1};
+    CHECK_HIP_ERROR(
+        hipMemcpy(bsr_row_ptr, local_ptr, sizeof(int) * (safe_size + 1), hipMemcpyHostToDevice));
+    CHECK_HIP_ERROR(
+        hipMemcpy(bsc_col_ptr, local_ptr, sizeof(int) * (safe_size + 1), hipMemcpyHostToDevice));
 
     size_t buffer_size;
     status = hipsparseXgebsr2gebsc_bufferSize<T>(nullptr,
@@ -436,312 +433,224 @@ void testing_gebsr2gebsc_bad_arg(void)
 template <typename T>
 hipsparseStatus_t testing_gebsr2gebsc(Arguments argus)
 {
-    hipsparseStatus_t status;
+    int                  m             = argus.M;
+    int                  n             = argus.N;
+    int                  row_block_dim = argus.row_block_dimA;
+    int                  col_block_dim = argus.col_block_dimA;
+    hipsparseAction_t    action        = argus.action;
+    hipsparseIndexBase_t base          = argus.baseA;
+    std::string          filename      = argus.filename;
 
     std::unique_ptr<handle_struct> unique_ptr_handle(new handle_struct);
     hipsparseHandle_t              handle = unique_ptr_handle->handle;
 
-    hipsparseAction_t    action = argus.action;
-    hipsparseIndexBase_t base   = argus.idx_base;
-
-    // Argument sanity check before allocating invalid memory
-    if((argus.M <= 0 && argus.M != -99) || (argus.N <= 0 && argus.M != -99)
-       || argus.row_block_dimA <= 0 || argus.col_block_dimA <= 0)
+    if(m == 0 || n == 0)
     {
-#if(defined(CUDART_VERSION))
-        if(argus.row_block_dimA == 0 || argus.col_block_dimA == 0)
-        {
-            return HIPSPARSE_STATUS_SUCCESS;
-        }
-#endif
-
-        int M             = argus.M;
-        int N             = argus.N;
-        int row_block_dim = argus.row_block_dimA;
-        int col_block_dim = argus.col_block_dimA;
-
-        static const size_t safe_size = 100;
-
-        // Allocate memory on device
-        DEVICE_ALLOC(T, dbuffer, safe_size);
-
-        if(!dbuffer)
-        {
-            return HIPSPARSE_STATUS_ALLOC_FAILED;
-        }
-
-        size_t buffer_size;
-
-        status = hipsparseXgebsr2gebsc_bufferSize<T>(handle,
-                                                     M,
-                                                     N,
-                                                     safe_size,
-                                                     (const T*)nullptr,
-                                                     nullptr,
-                                                     nullptr,
-                                                     row_block_dim,
-                                                     col_block_dim,
-                                                     &buffer_size);
-
-        if(M < 0 || N < 0 || row_block_dim <= 0 || col_block_dim <= 0)
-        {
-            verify_hipsparse_status_invalid_size(
-                status, "Error: M < 0 || N < 0 || row_block_dim < 0 || col_block_dim < 0");
-        }
-        else
-        {
-            verify_hipsparse_status_success(
-                status, "NOT(M < 0 || N < 0 || row_block_dim < 0 || col_block_dim < 0)");
-        }
-
-        status = hipsparseXgebsr2gebsc<T>(handle,
-                                          M,
-                                          N,
-                                          safe_size,
-                                          (const T*)nullptr,
-                                          nullptr,
-                                          nullptr,
-                                          row_block_dim,
-                                          col_block_dim,
-                                          (T*)nullptr,
-                                          nullptr,
-                                          nullptr,
-                                          action,
-                                          base,
-                                          dbuffer);
-
-        if(M < 0 || N < 0 || row_block_dim <= 0 || col_block_dim <= 0)
-        {
-            verify_hipsparse_status_invalid_size(
-                status, "Error: M < 0 || N < 0 || row_block_dim < 0 || col_block_dim < 0");
-        }
-        else
-        {
-            verify_hipsparse_status_success(
-                status, "NOT(M < 0 || N < 0 || row_block_dim < 0 || col_block_dim < 0)");
-        }
-
+#ifdef __HIP_PLATFORM_NVIDIA__
+        // cusparse does not support m == 0 for csr2bsr
         return HIPSPARSE_STATUS_SUCCESS;
+#endif
     }
 
-    //
-    // Build the gebsr matrix.
-    //
-    int              bsr_mb            = -1;
-    int              bsr_nb            = -1;
-    int              bsr_nnzb          = -1;
-    int              bsr_row_block_dim = -1;
-    int              bsr_col_block_dim = -1;
+    int mb = m * row_block_dim;
+    int nb = n * col_block_dim;
+
+    srand(12345ULL);
+
+    // Host structures
     std::vector<int> hbsr_row_ptr;
     std::vector<int> hbsr_col_ind;
     std::vector<T>   hbsr_val;
 
+    // Read or construct CSR matrix
+    int nnzb = 0;
+    if(!generate_csr_matrix(filename, mb, nb, nnzb, hbsr_row_ptr, hbsr_col_ind, hbsr_val, base))
     {
-        int m             = argus.M;
-        int n             = argus.N;
-        bsr_row_block_dim = argus.row_block_dimA;
-        bsr_col_block_dim = argus.col_block_dimA;
-
-        std::string binfile  = "";
-        std::string filename = "";
-
-        // When in testing mode, M == N == -99 indicates that we are testing with a real
-        // matrix from cise.ufl.edu
-        if(m == -99 && n == -99 && argus.timing == 0)
-        {
-            int safe_size = 100;
-            binfile       = argus.filename;
-            m = n = safe_size;
-        }
-
-        if(argus.timing == 1)
-        {
-            filename = argus.filename;
-        }
-
-        // Read or construct CSR matrix
-        std::vector<int> hcsr_row_ptr;
-        std::vector<int> hcsr_col_ind;
-        std::vector<T>   hcsr_val;
-        int              nnz;
-        srand(12345ULL);
-        if(binfile != "")
-        {
-            if(read_bin_matrix(
-                   binfile.c_str(), m, n, nnz, hcsr_row_ptr, hcsr_col_ind, hcsr_val, base)
-               != 0)
-            {
-                fprintf(stderr, "Cannot open [read] %s\n", binfile.c_str());
-                return HIPSPARSE_STATUS_INTERNAL_ERROR;
-            }
-        }
-        else if(argus.laplacian)
-        {
-            m = n = gen_2d_laplacian(argus.laplacian, hcsr_row_ptr, hcsr_col_ind, hcsr_val, base);
-            nnz   = hcsr_row_ptr[m];
-        }
-        else
-        {
-            std::vector<int> coo_row_ind;
-
-            if(filename != "")
-            {
-                if(read_mtx_matrix(
-                       filename.c_str(), m, n, nnz, coo_row_ind, hcsr_col_ind, hcsr_val, base)
-                   != 0)
-                {
-                    fprintf(stderr, "Cannot open [read] %s\n", filename.c_str());
-                    return HIPSPARSE_STATUS_INTERNAL_ERROR;
-                }
-            }
-            else
-            {
-                double scale = 0.02;
-                if(m > 1000 || n > 1000)
-                {
-                    scale = 2.0 / std::max(m, n);
-                }
-                nnz = m * scale * n;
-                nnz = std::max(nnz, 1);
-                gen_matrix_coo(m, n, nnz, coo_row_ind, hcsr_col_ind, hcsr_val, base);
-            }
-
-            // Convert COO to CSR
-            hcsr_row_ptr.resize(m + 1, 0);
-            for(int i = 0; i < nnz; ++i)
-            {
-                ++hcsr_row_ptr[coo_row_ind[i] + 1 - base];
-            }
-
-            hcsr_row_ptr[0] = base;
-            for(int i = 0; i < m; ++i)
-            {
-                hcsr_row_ptr[i + 1] += hcsr_row_ptr[i];
-            }
-        }
-
-        //
-        // Now convert the CSR matrix to a GEBSR matrix.
-        //
-        bsr_mb         = m;
-        bsr_nb         = n;
-        bsr_nnzb       = nnz;
-        size_t nvalues = bsr_nnzb * bsr_row_block_dim * bsr_col_block_dim;
-        hbsr_val.resize(nvalues);
-        for(size_t i = 0; i < nvalues; ++i)
-        {
-            hbsr_val[i] = random_generator<T>();
-        }
-        hbsr_row_ptr = hcsr_row_ptr;
-        hbsr_col_ind = hcsr_col_ind;
+        fprintf(stderr, "Cannot open [read] %s\ncol", filename.c_str());
+        return HIPSPARSE_STATUS_INTERNAL_ERROR;
     }
 
-    DEVICE_ALLOC(int, dbsr_row_ptr, (bsr_mb + 1));
-    DEVICE_ALLOC(int, dbsr_col_ind, bsr_nnzb);
-    DEVICE_ALLOC(T, dbsr_val, (bsr_nnzb * bsr_row_block_dim * bsr_col_block_dim));
+    m          = mb * row_block_dim;
+    n          = nb * col_block_dim;
+    size_t nnz = nnzb * row_block_dim * col_block_dim;
+
+    // Now use the csr matrix as the symbolic for the gebsr matrix.
+    hbsr_val.resize(nnz);
+    for(size_t i = 0; i < nnz; ++i)
+    {
+        hbsr_val[i] = random_generator<T>();
+    }
+
+    DEVICE_ALLOC(int, dbsr_row_ptr, (mb + 1));
+    DEVICE_ALLOC(int, dbsr_col_ind, nnzb);
+    DEVICE_ALLOC(T, dbsr_val, (nnzb * row_block_dim * col_block_dim));
 
     // Copy data from host to device
     CHECK_HIP_ERROR(hipMemcpy(
-        dbsr_row_ptr, hbsr_row_ptr.data(), sizeof(int) * (bsr_mb + 1), hipMemcpyHostToDevice));
-    CHECK_HIP_ERROR(hipMemcpy(
-        dbsr_col_ind, hbsr_col_ind.data(), sizeof(int) * bsr_nnzb, hipMemcpyHostToDevice));
+        dbsr_row_ptr, hbsr_row_ptr.data(), sizeof(int) * (mb + 1), hipMemcpyHostToDevice));
+    CHECK_HIP_ERROR(
+        hipMemcpy(dbsr_col_ind, hbsr_col_ind.data(), sizeof(int) * nnzb, hipMemcpyHostToDevice));
     CHECK_HIP_ERROR(hipMemcpy(dbsr_val,
                               hbsr_val.data(),
-                              sizeof(T) * bsr_nnzb * bsr_row_block_dim * bsr_col_block_dim,
+                              sizeof(T) * nnzb * row_block_dim * col_block_dim,
                               hipMemcpyHostToDevice));
 
-    //
     // Obtain required buffer size (from host)
-    //
     size_t buffer_size;
     CHECK_HIPSPARSE_ERROR(hipsparseXgebsr2gebsc_bufferSize<T>(handle,
-                                                              bsr_mb,
-                                                              bsr_nb,
-                                                              bsr_nnzb,
+                                                              mb,
+                                                              nb,
+                                                              nnzb,
                                                               dbsr_val,
                                                               dbsr_row_ptr,
                                                               dbsr_col_ind,
-                                                              bsr_row_block_dim,
-                                                              bsr_col_block_dim,
+                                                              row_block_dim,
+                                                              col_block_dim,
                                                               &buffer_size));
 
-    //
     // Allocate the buffer size.
-    //
     auto  dbuffer_managed = hipsparse_unique_ptr{device_malloc(buffer_size), device_free};
     void* dbuffer         = dbuffer_managed.get();
 
-    DEVICE_ALLOC(int, dbsc_row_ind, bsr_nnzb);
-    DEVICE_ALLOC(int, dbsc_col_ptr, (bsr_nb + 1));
-    DEVICE_ALLOC(T, dbsc_val, (bsr_nnzb * bsr_row_block_dim * bsr_col_block_dim));
+    DEVICE_ALLOC(int, dbsc_row_ind, nnzb);
+    DEVICE_ALLOC(int, dbsc_col_ptr, (nb + 1));
+    DEVICE_ALLOC(T, dbsc_val, (nnzb * row_block_dim * col_block_dim));
 
     if(argus.unit_check)
     {
         CHECK_HIPSPARSE_ERROR(hipsparseXgebsr2gebsc<T>(handle,
-                                                       bsr_mb,
-                                                       bsr_nb,
-                                                       bsr_nnzb,
+                                                       mb,
+                                                       nb,
+                                                       nnzb,
                                                        dbsr_val,
                                                        dbsr_row_ptr,
                                                        dbsr_col_ind,
-                                                       bsr_row_block_dim,
-                                                       bsr_col_block_dim,
+                                                       row_block_dim,
+                                                       col_block_dim,
                                                        dbsc_val,
                                                        dbsc_row_ind,
                                                        dbsc_col_ptr,
                                                        action,
                                                        base,
                                                        dbuffer));
-        //
         // Transfer to host.
-        //
-        std::vector<int> hbsc_from_device_row_ind(bsr_nnzb);
-        std::vector<int> hbsc_from_device_col_ptr(bsr_nb + 1);
-        std::vector<T>   hbsc_from_device_val(bsr_nnzb * bsr_row_block_dim * bsr_col_block_dim);
+        std::vector<int> hbsc_from_device_row_ind(nnzb);
+        std::vector<int> hbsc_from_device_col_ptr(nb + 1);
+        std::vector<T>   hbsc_from_device_val(nnzb * row_block_dim * col_block_dim);
         CHECK_HIP_ERROR(hipMemcpy(hbsc_from_device_col_ptr.data(),
                                   dbsc_col_ptr,
-                                  sizeof(int) * (bsr_nb + 1),
+                                  sizeof(int) * (nb + 1),
                                   hipMemcpyDeviceToHost));
         CHECK_HIP_ERROR(hipMemcpy(hbsc_from_device_row_ind.data(),
                                   dbsc_row_ind,
-                                  sizeof(int) * bsr_nnzb,
+                                  sizeof(int) * nnzb,
                                   hipMemcpyDeviceToHost));
         CHECK_HIP_ERROR(hipMemcpy(hbsc_from_device_val.data(),
                                   dbsc_val,
-                                  sizeof(T) * bsr_nnzb * bsr_row_block_dim * bsr_col_block_dim,
+                                  sizeof(T) * nnzb * row_block_dim * col_block_dim,
                                   hipMemcpyDeviceToHost));
 
-        //
         // Allocate host bsc matrix.
-        //
-        std::vector<int> hbsc_row_ind(bsr_nnzb);
-        std::vector<int> hbsc_col_ptr(bsr_nb + 1);
-        std::vector<T>   hbsc_val(bsr_nnzb * bsr_row_block_dim * bsr_col_block_dim);
-        host_gebsr_to_gebsc<T>(bsr_mb,
-                               bsr_nb,
-                               bsr_nnzb,
+        std::vector<int> hbsc_row_ind(nnzb);
+        std::vector<int> hbsc_col_ptr(nb + 1);
+        std::vector<T>   hbsc_val(nnzb * row_block_dim * col_block_dim);
+        host_gebsr_to_gebsc<T>(mb,
+                               nb,
+                               nnzb,
                                hbsr_row_ptr,
                                hbsr_col_ind,
                                hbsr_val,
-                               bsr_row_block_dim,
-                               bsr_col_block_dim,
+                               row_block_dim,
+                               col_block_dim,
                                hbsc_row_ind,
                                hbsc_col_ptr,
                                hbsc_val,
                                action,
                                base);
 
-        unit_check_general(1, bsr_nb + 1, 1, hbsc_from_device_col_ptr.data(), hbsc_col_ptr.data());
-        unit_check_general(1, bsr_nnzb, 1, hbsc_from_device_row_ind.data(), hbsc_row_ind.data());
+        unit_check_general(1, nb + 1, 1, hbsc_from_device_col_ptr.data(), hbsc_col_ptr.data());
+        unit_check_general(1, nnzb, 1, hbsc_from_device_row_ind.data(), hbsc_row_ind.data());
         if(action == HIPSPARSE_ACTION_NUMERIC)
         {
             unit_check_general(1,
-                               bsr_nnzb * bsr_row_block_dim * bsr_col_block_dim,
+                               nnzb * row_block_dim * col_block_dim,
                                1,
                                hbsc_from_device_val.data(),
                                hbsc_val.data());
         }
     }
+
+    if(argus.timing)
+    {
+        int number_cold_calls = 2;
+        int number_hot_calls  = argus.iters;
+
+        // Warm up
+        for(int iter = 0; iter < number_cold_calls; ++iter)
+        {
+            CHECK_HIPSPARSE_ERROR(hipsparseXgebsr2gebsc<T>(handle,
+                                                           mb,
+                                                           nb,
+                                                           nnzb,
+                                                           dbsr_val,
+                                                           dbsr_row_ptr,
+                                                           dbsr_col_ind,
+                                                           row_block_dim,
+                                                           col_block_dim,
+                                                           dbsc_val,
+                                                           dbsc_row_ind,
+                                                           dbsc_col_ptr,
+                                                           action,
+                                                           base,
+                                                           dbuffer));
+        }
+
+        double gpu_time_used = get_time_us();
+
+        // Performance run
+        for(int iter = 0; iter < number_hot_calls; ++iter)
+        {
+            CHECK_HIPSPARSE_ERROR(hipsparseXgebsr2gebsc<T>(handle,
+                                                           mb,
+                                                           nb,
+                                                           nnzb,
+                                                           dbsr_val,
+                                                           dbsr_row_ptr,
+                                                           dbsr_col_ind,
+                                                           row_block_dim,
+                                                           col_block_dim,
+                                                           dbsc_val,
+                                                           dbsc_row_ind,
+                                                           dbsc_col_ptr,
+                                                           action,
+                                                           base,
+                                                           dbuffer));
+        }
+
+        gpu_time_used = (get_time_us() - gpu_time_used) / number_hot_calls;
+
+        double gbyte_count
+            = gebsr2gebsc_gbyte_count<T>(mb, nb, nnzb, row_block_dim, col_block_dim, action);
+        double gpu_gbyte = get_gpu_gbyte(gpu_time_used, gbyte_count);
+
+        display_timing_info(display_key_t::Mb,
+                            mb,
+                            display_key_t::Nb,
+                            nb,
+                            display_key_t::nnzb,
+                            nnzb,
+                            display_key_t::row_block_dim,
+                            row_block_dim,
+                            display_key_t::col_block_dim,
+                            col_block_dim,
+                            display_key_t::action,
+                            hipsparse_action2string(action),
+                            display_key_t::bandwidth,
+                            gpu_gbyte,
+                            display_key_t::time_ms,
+                            get_gpu_time_msec(gpu_time_used));
+    }
+
     return HIPSPARSE_STATUS_SUCCESS;
 }
 
